from PIL.Image import (
    FLIP_LEFT_RIGHT,
    FLIP_TOP_BOTTOM,
    ROTATE_90,
    ROTATE_180,
    ROTATE_270,
    TRANSPOSE,
    TRANSVERSE,
)

from . import helper
from .helper import assert_image_equal

HOPPER = {
    mode: helper.hopper(mode).crop((0, 0, 121, 127)).copy()
    for mode in ["L", "RGB", "I;16", "I;16L", "I;16B"]
}


def test_flip_left_right():
    def transpose(mode):
        im = HOPPER[mode]
        out = im.transpose(FLIP_LEFT_RIGHT)
        assert out.mode == mode
        assert out.size == im.size

        x, y = im.size
        assert im.getpixel((1, 1)) == out.getpixel((x - 2, 1))
        assert im.getpixel((x - 2, 1)) == out.getpixel((1, 1))
        assert im.getpixel((1, y - 2)) == out.getpixel((x - 2, y - 2))
        assert im.getpixel((x - 2, y - 2)) == out.getpixel((1, y - 2))

    for mode in HOPPER:
        transpose(mode)


def test_flip_top_bottom():
    def transpose(mode):
        im = HOPPER[mode]
        out = im.transpose(FLIP_TOP_BOTTOM)
        assert out.mode == mode
        assert out.size == im.size

        x, y = im.size
        assert im.getpixel((1, 1)) == out.getpixel((1, y - 2))
        assert im.getpixel((x - 2, 1)) == out.getpixel((x - 2, y - 2))
        assert im.getpixel((1, y - 2)) == out.getpixel((1, 1))
        assert im.getpixel((x - 2, y - 2)) == out.getpixel((x - 2, 1))

    for mode in HOPPER:
        transpose(mode)


def test_rotate_90():
    def transpose(mode):
        im = HOPPER[mode]
        out = im.transpose(ROTATE_90)
        assert out.mode == mode
        assert out.size == im.size[::-1]

        x, y = im.size
        assert im.getpixel((1, 1)) == out.getpixel((1, x - 2))
        assert im.getpixel((x - 2, 1)) == out.getpixel((1, 1))
        assert im.getpixel((1, y - 2)) == out.getpixel((y - 2, x - 2))
        assert im.getpixel((x - 2, y - 2)) == out.getpixel((y - 2, 1))

    for mode in HOPPER:
        transpose(mode)


def test_rotate_180():
    def transpose(mode):
        im = HOPPER[mode]
        out = im.transpose(ROTATE_180)
        assert out.mode == mode
        assert out.size == im.size

        x, y = im.size
        assert im.getpixel((1, 1)) == out.getpixel((x - 2, y - 2))
        assert im.getpixel((x - 2, 1)) == out.getpixel((1, y - 2))
        assert im.getpixel((1, y - 2)) == out.getpixel((x - 2, 1))
        assert im.getpixel((x - 2, y - 2)) == out.getpixel((1, 1))

    for mode in HOPPER:
        transpose(mode)


def test_rotate_270():
    def transpose(mode):
        im = HOPPER[mode]
        out = im.transpose(ROTATE_270)
        assert out.mode == mode
        assert out.size == im.size[::-1]

        x, y = im.size
        assert im.getpixel((1, 1)) == out.getpixel((y - 2, 1))
        assert im.getpixel((x - 2, 1)) == out.getpixel((y - 2, x - 2))
        assert im.getpixel((1, y - 2)) == out.getpixel((1, 1))
        assert im.getpixel((x - 2, y - 2)) == out.getpixel((1, x - 2))

    for mode in HOPPER:
        transpose(mode)


def test_transpose():
    def transpose(mode):
        im = HOPPER[mode]
        out = im.transpose(TRANSPOSE)
        assert out.mode == mode
        assert out.size == im.size[::-1]

        x, y = im.size
        assert im.getpixel((1, 1)) == out.getpixel((1, 1))
        assert im.getpixel((x - 2, 1)) == out.getpixel((1, x - 2))
        assert im.getpixel((1, y - 2)) == out.getpixel((y - 2, 1))
        assert im.getpixel((x - 2, y - 2)) == out.getpixel((y - 2, x - 2))

    for mode in HOPPER:
        transpose(mode)


def test_tranverse():
    def transpose(mode):
        im = HOPPER[mode]
        out = im.transpose(TRANSVERSE)
        assert out.mode == mode
        assert out.size == im.size[::-1]

        x, y = im.size
        assert im.getpixel((1, 1)) == out.getpixel((y - 2, x - 2))
        assert im.getpixel((x - 2, 1)) == out.getpixel((y - 2, 1))
        assert im.getpixel((1, y - 2)) == out.getpixel((1, x - 2))
        assert im.getpixel((x - 2, y - 2)) == out.getpixel((1, 1))

    for mode in HOPPER:
        transpose(mode)


def test_roundtrip():
    for mode in HOPPER:
        im = HOPPER[mode]

        def transpose(first, second):
            return im.transpose(first).transpose(second)

        assert_image_equal(im, transpose(FLIP_LEFT_RIGHT, FLIP_LEFT_RIGHT))
        assert_image_equal(im, transpose(FLIP_TOP_BOTTOM, FLIP_TOP_BOTTOM))
        assert_image_equal(im, transpose(ROTATE_90, ROTATE_270))
        assert_image_equal(im, transpose(ROTATE_180, ROTATE_180))
        assert_image_equal(
            im.transpose(TRANSPOSE), transpose(ROTATE_90, FLIP_TOP_BOTTOM)
        )
        assert_image_equal(
            im.transpose(TRANSPOSE), transpose(ROTATE_270, FLIP_LEFT_RIGHT)
        )
        assert_image_equal(
            im.transpose(TRANSVERSE), transpose(ROTATE_90, FLIP_LEFT_RIGHT)
        )
        assert_image_equal(
            im.transpose(TRANSVERSE), transpose(ROTATE_270, FLIP_TOP_BOTTOM)
        )
        assert_image_equal(im.transpose(TRANSVERSE), transpose(ROTATE_180, TRANSPOSE))



import io

import pytest

from PIL import BmpImagePlugin, Image

from .helper import assert_image_equal, assert_image_equal_tofile, hopper


def test_sanity(tmp_path):
    def roundtrip(im):
        outfile = str(tmp_path / "temp.bmp")

        im.save(outfile, "BMP")

        with Image.open(outfile) as reloaded:
            reloaded.load()
            assert im.mode == reloaded.mode
            assert im.size == reloaded.size
            assert reloaded.format == "BMP"
            assert reloaded.get_format_mimetype() == "image/bmp"

    roundtrip(hopper())

    roundtrip(hopper("1"))
    roundtrip(hopper("L"))
    roundtrip(hopper("P"))
    roundtrip(hopper("RGB"))


def test_invalid_file():
    with open("Tests/images/flower.jpg", "rb") as fp:
        with pytest.raises(SyntaxError):
            BmpImagePlugin.BmpImageFile(fp)


def test_save_to_bytes():
    output = io.BytesIO()
    im = hopper()
    im.save(output, "BMP")

    output.seek(0)
    with Image.open(output) as reloaded:
        assert im.mode == reloaded.mode
        assert im.size == reloaded.size
        assert reloaded.format == "BMP"


def test_save_too_large(tmp_path):
    outfile = str(tmp_path / "temp.bmp")
    with Image.new("RGB", (1, 1)) as im:
        im._size = (37838, 37838)
        with pytest.raises(ValueError):
            im.save(outfile)


def test_dpi():
    dpi = (72, 72)

    output = io.BytesIO()
    with hopper() as im:
        im.save(output, "BMP", dpi=dpi)

    output.seek(0)
    with Image.open(output) as reloaded:
        assert reloaded.info["dpi"] == (72.008961115161, 72.008961115161)


def test_save_bmp_with_dpi(tmp_path):
    # Test for #1301
    # Arrange
    outfile = str(tmp_path / "temp.jpg")
    with Image.open("Tests/images/hopper.bmp") as im:
        assert im.info["dpi"] == (95.98654816726399, 95.98654816726399)

        # Act
        im.save(outfile, "JPEG", dpi=im.info["dpi"])

        # Assert
        with Image.open(outfile) as reloaded:
            reloaded.load()
            assert reloaded.info["dpi"] == (96, 96)
            assert reloaded.size == im.size
            assert reloaded.format == "JPEG"


def test_save_float_dpi(tmp_path):
    outfile = str(tmp_path / "temp.bmp")
    with Image.open("Tests/images/hopper.bmp") as im:
        im.save(outfile, dpi=(72.21216100543306, 72.21216100543306))
        with Image.open(outfile) as reloaded:
            assert reloaded.info["dpi"] == (72.21216100543306, 72.21216100543306)


def test_load_dib():
    # test for #1293, Imagegrab returning Unsupported Bitfields Format
    with Image.open("Tests/images/clipboard.dib") as im:
        assert im.format == "DIB"
        assert im.get_format_mimetype() == "image/bmp"

        assert_image_equal_tofile(im, "Tests/images/clipboard_target.png")


def test_save_dib(tmp_path):
    outfile = str(tmp_path / "temp.dib")

    with Image.open("Tests/images/clipboard.dib") as im:
        im.save(outfile)

        with Image.open(outfile) as reloaded:
            assert reloaded.format == "DIB"
            assert reloaded.get_format_mimetype() == "image/bmp"
            assert_image_equal(im, reloaded)


def test_rgba_bitfields():
    # This test image has been manually hexedited
    # to change the bitfield compression in the header from XBGR to RGBA
    with Image.open("Tests/images/rgb32bf-rgba.bmp") as im:

        # So before the comparing the image, swap the channels
        b, g, r = im.split()[1:]
        im = Image.merge("RGB", (r, g, b))

    assert_image_equal_tofile(im, "Tests/images/bmp/q/rgb32bf-xbgr.bmp")


def test_offset():
    # This image has been hexedited
    # to exclude the palette size from the pixel data offset
    with Image.open("Tests/images/pal8_offset.bmp") as im:
        assert_image_equal_tofile(im, "Tests/images/bmp/g/pal8.bmp")



import os
from glob import glob
from itertools import product

import pytest

from PIL import Image

from .helper import assert_image_equal, assert_image_equal_tofile, hopper

_TGA_DIR = os.path.join("Tests", "images", "tga")
_TGA_DIR_COMMON = os.path.join(_TGA_DIR, "common")


_MODES = ("L", "LA", "P", "RGB", "RGBA")
_ORIGINS = ("tl", "bl")

_ORIGIN_TO_ORIENTATION = {"tl": 1, "bl": -1}


def test_sanity(tmp_path):
    for mode in _MODES:

        def roundtrip(original_im):
            out = str(tmp_path / "temp.tga")

            original_im.save(out, rle=rle)
            with Image.open(out) as saved_im:
                if rle:
                    assert (
                        saved_im.info["compression"] == original_im.info["compression"]
                    )
                assert saved_im.info["orientation"] == original_im.info["orientation"]
                if mode == "P":
                    assert saved_im.getpalette() == original_im.getpalette()

                assert_image_equal(saved_im, original_im)

        png_paths = glob(os.path.join(_TGA_DIR_COMMON, f"*x*_{mode.lower()}.png"))

        for png_path in png_paths:
            with Image.open(png_path) as reference_im:
                assert reference_im.mode == mode

                path_no_ext = os.path.splitext(png_path)[0]
                for origin, rle in product(_ORIGINS, (True, False)):
                    tga_path = "{}_{}_{}.tga".format(
                        path_no_ext, origin, "rle" if rle else "raw"
                    )

                    with Image.open(tga_path) as original_im:
                        assert original_im.format == "TGA"
                        assert original_im.get_format_mimetype() == "image/x-tga"
                        if rle:
                            assert original_im.info["compression"] == "tga_rle"
                        assert (
                            original_im.info["orientation"]
                            == _ORIGIN_TO_ORIENTATION[origin]
                        )
                        if mode == "P":
                            assert original_im.getpalette() == reference_im.getpalette()

                        assert_image_equal(original_im, reference_im)

                        roundtrip(original_im)


def test_palette_depth_16(tmp_path):
    with Image.open("Tests/images/p_16.tga") as im:
        assert_image_equal_tofile(im.convert("RGB"), "Tests/images/p_16.png")

        out = str(tmp_path / "temp.png")
        im.save(out)
        with Image.open(out) as reloaded:
            assert_image_equal_tofile(reloaded.convert("RGB"), "Tests/images/p_16.png")


def test_id_field():
    # tga file with id field
    test_file = "Tests/images/tga_id_field.tga"

    # Act
    with Image.open(test_file) as im:

        # Assert
        assert im.size == (100, 100)


def test_id_field_rle():
    # tga file with id field
    test_file = "Tests/images/rgb32rle.tga"

    # Act
    with Image.open(test_file) as im:

        # Assert
        assert im.size == (199, 199)


def test_save(tmp_path):
    test_file = "Tests/images/tga_id_field.tga"
    with Image.open(test_file) as im:
        out = str(tmp_path / "temp.tga")

        # Save
        im.save(out)
        with Image.open(out) as test_im:
            assert test_im.size == (100, 100)
            assert test_im.info["id_section"] == im.info["id_section"]

        # RGBA save
        im.convert("RGBA").save(out)
    with Image.open(out) as test_im:
        assert test_im.size == (100, 100)


def test_save_wrong_mode(tmp_path):
    im = hopper("PA")
    out = str(tmp_path / "temp.tga")

    with pytest.raises(OSError):
        im.save(out)


def test_save_mapdepth():
    # This image has been manually hexedited from 200x32_p_bl_raw.tga
    # to include an origin
    test_file = "Tests/images/200x32_p_bl_raw_origin.tga"
    with Image.open(test_file) as im:
        assert_image_equal_tofile(im, "Tests/images/tga/common/200x32_p.png")


def test_save_id_section(tmp_path):
    test_file = "Tests/images/rgb32rle.tga"
    with Image.open(test_file) as im:
        out = str(tmp_path / "temp.tga")

        # Check there is no id section
        im.save(out)
    with Image.open(out) as test_im:
        assert "id_section" not in test_im.info

    # Save with custom id section
    im.save(out, id_section=b"Test content")
    with Image.open(out) as test_im:
        assert test_im.info["id_section"] == b"Test content"

    # Save with custom id section greater than 255 characters
    id_section = b"Test content" * 25
    pytest.warns(UserWarning, lambda: im.save(out, id_section=id_section))
    with Image.open(out) as test_im:
        assert test_im.info["id_section"] == id_section[:255]

    test_file = "Tests/images/tga_id_field.tga"
    with Image.open(test_file) as im:

        # Save with no id section
        im.save(out, id_section="")
    with Image.open(out) as test_im:
        assert "id_section" not in test_im.info


def test_save_orientation(tmp_path):
    test_file = "Tests/images/rgb32rle.tga"
    out = str(tmp_path / "temp.tga")
    with Image.open(test_file) as im:
        assert im.info["orientation"] == -1

        im.save(out, orientation=1)
    with Image.open(out) as test_im:
        assert test_im.info["orientation"] == 1


def test_horizontal_orientations():
    # These images have been manually hexedited to have the relevant orientations
    with Image.open("Tests/images/rgb32rle_top_right.tga") as im:
        assert im.load()[90, 90][:3] == (0, 0, 0)

    with Image.open("Tests/images/rgb32rle_bottom_right.tga") as im:
        assert im.load()[90, 90][:3] == (0, 255, 0)


def test_save_rle(tmp_path):
    test_file = "Tests/images/rgb32rle.tga"
    with Image.open(test_file) as im:
        assert im.info["compression"] == "tga_rle"

        out = str(tmp_path / "temp.tga")

        # Save
        im.save(out)
    with Image.open(out) as test_im:
        assert test_im.size == (199, 199)
        assert test_im.info["compression"] == "tga_rle"

    # Save without compression
    im.save(out, compression=None)
    with Image.open(out) as test_im:
        assert "compression" not in test_im.info

    # RGBA save
    im.convert("RGBA").save(out)
    with Image.open(out) as test_im:
        assert test_im.size == (199, 199)

    test_file = "Tests/images/tga_id_field.tga"
    with Image.open(test_file) as im:
        assert "compression" not in im.info

        # Save with compression
        im.save(out, compression="tga_rle")
    with Image.open(out) as test_im:
        assert test_im.info["compression"] == "tga_rle"


def test_save_l_transparency(tmp_path):
    # There are 559 transparent pixels in la.tga.
    num_transparent = 559

    in_file = "Tests/images/la.tga"
    with Image.open(in_file) as im:
        assert im.mode == "LA"
        assert im.getchannel("A").getcolors()[0][0] == num_transparent

        out = str(tmp_path / "temp.tga")
        im.save(out)

    with Image.open(out) as test_im:
        assert test_im.mode == "LA"
        assert test_im.getchannel("A").getcolors()[0][0] == num_transparent

        assert_image_equal(im, test_im)



# Test the ImageMorphology functionality
import pytest

from PIL import Image, ImageMorph, _imagingmorph

from .helper import assert_image_equal_tofile, hopper


def string_to_img(image_string):
    """Turn a string image representation into a binary image"""
    rows = [s for s in image_string.replace(" ", "").split("\n") if len(s)]
    height = len(rows)
    width = len(rows[0])
    im = Image.new("L", (width, height))
    for i in range(width):
        for j in range(height):
            c = rows[j][i]
            v = c in "X1"
            im.putpixel((i, j), v)

    return im


A = string_to_img(
    """
    .......
    .......
    ..111..
    ..111..
    ..111..
    .......
    .......
    """
)


def img_to_string(im):
    """Turn a (small) binary image into a string representation"""
    chars = ".1"
    width, height = im.size
    return "\n".join(
        "".join(chars[im.getpixel((c, r)) > 0] for c in range(width))
        for r in range(height)
    )


def img_string_normalize(im):
    return img_to_string(string_to_img(im))


def assert_img_equal(A, B):
    assert img_to_string(A) == img_to_string(B)


def assert_img_equal_img_string(A, Bstring):
    assert img_to_string(A) == img_string_normalize(Bstring)


def test_str_to_img():
    assert_image_equal_tofile(A, "Tests/images/morph_a.png")


def create_lut():
    for op in ("corner", "dilation4", "dilation8", "erosion4", "erosion8", "edge"):
        lb = ImageMorph.LutBuilder(op_name=op)
        lut = lb.build_lut()
        with open(f"Tests/images/{op}.lut", "wb") as f:
            f.write(lut)


# create_lut()
def test_lut():
    for op in ("corner", "dilation4", "dilation8", "erosion4", "erosion8", "edge"):
        lb = ImageMorph.LutBuilder(op_name=op)
        assert lb.get_lut() is None

        lut = lb.build_lut()
        with open(f"Tests/images/{op}.lut", "rb") as f:
            assert lut == bytearray(f.read())


def test_no_operator_loaded():
    mop = ImageMorph.MorphOp()
    with pytest.raises(Exception) as e:
        mop.apply(None)
    assert str(e.value) == "No operator loaded"
    with pytest.raises(Exception) as e:
        mop.match(None)
    assert str(e.value) == "No operator loaded"
    with pytest.raises(Exception) as e:
        mop.save_lut(None)
    assert str(e.value) == "No operator loaded"


# Test the named patterns
def test_erosion8():
    # erosion8
    mop = ImageMorph.MorphOp(op_name="erosion8")
    count, Aout = mop.apply(A)
    assert count == 8
    assert_img_equal_img_string(
        Aout,
        """
                                     .......
                                     .......
                                     .......
                                     ...1...
                                     .......
                                     .......
                                     .......
                                     """,
    )


def test_dialation8():
    # dialation8
    mop = ImageMorph.MorphOp(op_name="dilation8")
    count, Aout = mop.apply(A)
    assert count == 16
    assert_img_equal_img_string(
        Aout,
        """
                                     .......
                                     .11111.
                                     .11111.
                                     .11111.
                                     .11111.
                                     .11111.
                                     .......
                                     """,
    )


def test_erosion4():
    # erosion4
    mop = ImageMorph.MorphOp(op_name="dilation4")
    count, Aout = mop.apply(A)
    assert count == 12
    assert_img_equal_img_string(
        Aout,
        """
                                     .......
                                     ..111..
                                     .11111.
                                     .11111.
                                     .11111.
                                     ..111..
                                     .......
                                     """,
    )


def test_edge():
    # edge
    mop = ImageMorph.MorphOp(op_name="edge")
    count, Aout = mop.apply(A)
    assert count == 1
    assert_img_equal_img_string(
        Aout,
        """
                                     .......
                                     .......
                                     ..111..
                                     ..1.1..
                                     ..111..
                                     .......
                                     .......
                                     """,
    )


def test_corner():
    # Create a corner detector pattern
    mop = ImageMorph.MorphOp(patterns=["1:(... ... ...)->0", "4:(00. 01. ...)->1"])
    count, Aout = mop.apply(A)
    assert count == 5
    assert_img_equal_img_string(
        Aout,
        """
                                     .......
                                     .......
                                     ..1.1..
                                     .......
                                     ..1.1..
                                     .......
                                     .......
                                     """,
    )

    # Test the coordinate counting with the same operator
    coords = mop.match(A)
    assert len(coords) == 4
    assert tuple(coords) == ((2, 2), (4, 2), (2, 4), (4, 4))

    coords = mop.get_on_pixels(Aout)
    assert len(coords) == 4
    assert tuple(coords) == ((2, 2), (4, 2), (2, 4), (4, 4))


def test_mirroring():
    # Test 'M' for mirroring
    mop = ImageMorph.MorphOp(patterns=["1:(... ... ...)->0", "M:(00. 01. ...)->1"])
    count, Aout = mop.apply(A)
    assert count == 7
    assert_img_equal_img_string(
        Aout,
        """
                                     .......
                                     .......
                                     ..1.1..
                                     .......
                                     .......
                                     .......
                                     .......
                                     """,
    )


def test_negate():
    # Test 'N' for negate
    mop = ImageMorph.MorphOp(patterns=["1:(... ... ...)->0", "N:(00. 01. ...)->1"])
    count, Aout = mop.apply(A)
    assert count == 8
    assert_img_equal_img_string(
        Aout,
        """
                                     .......
                                     .......
                                     ..1....
                                     .......
                                     .......
                                     .......
                                     .......
                                     """,
    )


def test_incorrect_mode():
    im = hopper("RGB")
    mop = ImageMorph.MorphOp(op_name="erosion8")

    with pytest.raises(ValueError) as e:
        mop.apply(im)
    assert str(e.value) == "Image mode must be L"
    with pytest.raises(ValueError) as e:
        mop.match(im)
    assert str(e.value) == "Image mode must be L"
    with pytest.raises(ValueError) as e:
        mop.get_on_pixels(im)
    assert str(e.value) == "Image mode must be L"


def test_add_patterns():
    # Arrange
    lb = ImageMorph.LutBuilder(op_name="corner")
    assert lb.patterns == ["1:(... ... ...)->0", "4:(00. 01. ...)->1"]
    new_patterns = ["M:(00. 01. ...)->1", "N:(00. 01. ...)->1"]

    # Act
    lb.add_patterns(new_patterns)

    # Assert
    assert lb.patterns == [
        "1:(... ... ...)->0",
        "4:(00. 01. ...)->1",
        "M:(00. 01. ...)->1",
        "N:(00. 01. ...)->1",
    ]


def test_unknown_pattern():
    with pytest.raises(Exception):
        ImageMorph.LutBuilder(op_name="unknown")


def test_pattern_syntax_error():
    # Arrange
    lb = ImageMorph.LutBuilder(op_name="corner")
    new_patterns = ["a pattern with a syntax error"]
    lb.add_patterns(new_patterns)

    # Act / Assert
    with pytest.raises(Exception) as e:
        lb.build_lut()
    assert str(e.value) == 'Syntax error in pattern "a pattern with a syntax error"'


def test_load_invalid_mrl():
    # Arrange
    invalid_mrl = "Tests/images/hopper.png"
    mop = ImageMorph.MorphOp()

    # Act / Assert
    with pytest.raises(Exception) as e:
        mop.load_lut(invalid_mrl)
    assert str(e.value) == "Wrong size operator file!"


def test_roundtrip_mrl(tmp_path):
    # Arrange
    tempfile = str(tmp_path / "temp.mrl")
    mop = ImageMorph.MorphOp(op_name="corner")
    initial_lut = mop.lut

    # Act
    mop.save_lut(tempfile)
    mop.load_lut(tempfile)

    # Act / Assert
    assert mop.lut == initial_lut


def test_set_lut():
    # Arrange
    lb = ImageMorph.LutBuilder(op_name="corner")
    lut = lb.build_lut()
    mop = ImageMorph.MorphOp()

    # Act
    mop.set_lut(lut)

    # Assert
    assert mop.lut == lut


def test_wrong_mode():
    lut = ImageMorph.LutBuilder(op_name="corner").build_lut()
    imrgb = Image.new("RGB", (10, 10))
    iml = Image.new("L", (10, 10))

    with pytest.raises(RuntimeError):
        _imagingmorph.apply(bytes(lut), imrgb.im.id, iml.im.id)

    with pytest.raises(RuntimeError):
        _imagingmorph.apply(bytes(lut), iml.im.id, imrgb.im.id)

    with pytest.raises(RuntimeError):
        _imagingmorph.match(bytes(lut), imrgb.im.id)

    # Should not raise
    _imagingmorph.match(bytes(lut), iml.im.id)



import sys

import pytest

from PIL import Image

# This test is not run automatically.
#
# It requires > 2gb memory for the >2 gigapixel image generated in the
# second test.  Running this automatically would amount to a denial of
# service on our testing infrastructure.  I expect this test to fail
# on any 32-bit machine, as well as any smallish things (like
# Raspberry Pis). It does succeed on a 3gb Ubuntu 12.04x64 VM on Python
# 2.7 and 3.2.


try:
    import numpy
except ImportError:
    numpy = None

YDIM = 32769
XDIM = 48000


pytestmark = pytest.mark.skipif(sys.maxsize <= 2 ** 32, reason="requires 64-bit system")


def _write_png(tmp_path, xdim, ydim):
    f = str(tmp_path / "temp.png")
    im = Image.new("L", (xdim, ydim), 0)
    im.save(f)


def test_large(tmp_path):
    """succeeded prepatch"""
    _write_png(tmp_path, XDIM, YDIM)


def test_2gpx(tmp_path):
    """failed prepatch"""
    _write_png(tmp_path, XDIM, XDIM)


@pytest.mark.skipif(numpy is None, reason="Numpy is not installed")
def test_size_greater_than_int():
    arr = numpy.ndarray(shape=(16394, 16394))
    Image.fromarray(arr)



import pytest

from PIL import Image

from .helper import assert_image_similar, hopper, is_ppc64le


def test_sanity():
    image = hopper()
    converted = image.quantize()
    assert converted.mode == "P"
    assert_image_similar(converted.convert("RGB"), image, 10)

    image = hopper()
    converted = image.quantize(palette=hopper("P"))
    assert converted.mode == "P"
    assert_image_similar(converted.convert("RGB"), image, 60)


@pytest.mark.xfail(is_ppc64le(), reason="failing on ppc64le on GHA")
def test_libimagequant_quantize():
    image = hopper()
    try:
        converted = image.quantize(100, Image.LIBIMAGEQUANT)
    except ValueError as ex:  # pragma: no cover
        if "dependency" in str(ex).lower():
            pytest.skip("libimagequant support not available")
        else:
            raise
    assert converted.mode == "P"
    assert_image_similar(converted.convert("RGB"), image, 15)
    assert len(converted.getcolors()) == 100


def test_octree_quantize():
    image = hopper()
    converted = image.quantize(100, Image.FASTOCTREE)
    assert converted.mode == "P"
    assert_image_similar(converted.convert("RGB"), image, 20)
    assert len(converted.getcolors()) == 100


def test_rgba_quantize():
    image = hopper("RGBA")
    with pytest.raises(ValueError):
        image.quantize(method=0)

    assert image.quantize().convert().mode == "RGBA"


def test_quantize():
    with Image.open("Tests/images/caption_6_33_22.png") as image:
        image = image.convert("RGB")
    converted = image.quantize()
    assert converted.mode == "P"
    assert_image_similar(converted.convert("RGB"), image, 1)


def test_quantize_no_dither():
    image = hopper()
    with Image.open("Tests/images/caption_6_33_22.png") as palette:
        palette = palette.convert("P")

    converted = image.quantize(dither=0, palette=palette)
    assert converted.mode == "P"
    assert converted.palette.palette == palette.palette.palette


def test_quantize_dither_diff():
    image = hopper()
    with Image.open("Tests/images/caption_6_33_22.png") as palette:
        palette = palette.convert("P")

    dither = image.quantize(dither=1, palette=palette)
    nodither = image.quantize(dither=0, palette=palette)

    assert dither.tobytes() != nodither.tobytes()


def test_colors():
    im = hopper()
    colors = 2
    converted = im.quantize(colors)
    assert len(converted.palette.palette) == colors * len("RGB")


def test_transparent_colors_equal():
    im = Image.new("RGBA", (1, 2), (0, 0, 0, 0))
    px = im.load()
    px[0, 1] = (255, 255, 255, 0)

    converted = im.quantize()
    converted_px = converted.load()
    assert converted_px[0, 0] == converted_px[0, 1]


@pytest.mark.parametrize(
    "method, color",
    (
        (Image.MEDIANCUT, (0, 0, 0)),
        (Image.MAXCOVERAGE, (0, 0, 0)),
        (Image.FASTOCTREE, (0, 0, 0)),
        (Image.FASTOCTREE, (0, 0, 0, 0)),
    ),
)
def test_palette(method, color):
    im = Image.new("RGBA" if len(color) == 4 else "RGB", (1, 1), color)

    converted = im.quantize(method=method)
    converted_px = converted.load()
    assert converted_px[0, 0] == converted.palette.colors[color]



import pytest

from PIL import Image, ImageFilter

from .helper import assert_image_equal, hopper


def test_sanity():
    def apply_filter(filter_to_apply):
        for mode in ["L", "RGB", "CMYK"]:
            im = hopper(mode)
            out = im.filter(filter_to_apply)
            assert out.mode == im.mode
            assert out.size == im.size

    apply_filter(ImageFilter.BLUR)
    apply_filter(ImageFilter.CONTOUR)
    apply_filter(ImageFilter.DETAIL)
    apply_filter(ImageFilter.EDGE_ENHANCE)
    apply_filter(ImageFilter.EDGE_ENHANCE_MORE)
    apply_filter(ImageFilter.EMBOSS)
    apply_filter(ImageFilter.FIND_EDGES)
    apply_filter(ImageFilter.SMOOTH)
    apply_filter(ImageFilter.SMOOTH_MORE)
    apply_filter(ImageFilter.SHARPEN)
    apply_filter(ImageFilter.MaxFilter)
    apply_filter(ImageFilter.MedianFilter)
    apply_filter(ImageFilter.MinFilter)
    apply_filter(ImageFilter.ModeFilter)
    apply_filter(ImageFilter.GaussianBlur)
    apply_filter(ImageFilter.GaussianBlur(5))
    apply_filter(ImageFilter.BoxBlur(5))
    apply_filter(ImageFilter.UnsharpMask)
    apply_filter(ImageFilter.UnsharpMask(10))

    with pytest.raises(TypeError):
        apply_filter("hello")


def test_crash():

    # crashes on small images
    im = Image.new("RGB", (1, 1))
    im.filter(ImageFilter.SMOOTH)

    im = Image.new("RGB", (2, 2))
    im.filter(ImageFilter.SMOOTH)

    im = Image.new("RGB", (3, 3))
    im.filter(ImageFilter.SMOOTH)


def test_modefilter():
    def modefilter(mode):
        im = Image.new(mode, (3, 3), None)
        im.putdata(list(range(9)))
        # image is:
        #   0 1 2
        #   3 4 5
        #   6 7 8
        mod = im.filter(ImageFilter.ModeFilter).getpixel((1, 1))
        im.putdata([0, 0, 1, 2, 5, 1, 5, 2, 0])  # mode=0
        mod2 = im.filter(ImageFilter.ModeFilter).getpixel((1, 1))
        return mod, mod2

    assert modefilter("1") == (4, 0)
    assert modefilter("L") == (4, 0)
    assert modefilter("P") == (4, 0)
    assert modefilter("RGB") == ((4, 0, 0), (0, 0, 0))


def test_rankfilter():
    def rankfilter(mode):
        im = Image.new(mode, (3, 3), None)
        im.putdata(list(range(9)))
        # image is:
        #   0 1 2
        #   3 4 5
        #   6 7 8
        minimum = im.filter(ImageFilter.MinFilter).getpixel((1, 1))
        med = im.filter(ImageFilter.MedianFilter).getpixel((1, 1))
        maximum = im.filter(ImageFilter.MaxFilter).getpixel((1, 1))
        return minimum, med, maximum

    assert rankfilter("1") == (0, 4, 8)
    assert rankfilter("L") == (0, 4, 8)
    with pytest.raises(ValueError):
        rankfilter("P")
    assert rankfilter("RGB") == ((0, 0, 0), (4, 0, 0), (8, 0, 0))
    assert rankfilter("I") == (0, 4, 8)
    assert rankfilter("F") == (0.0, 4.0, 8.0)


def test_rankfilter_properties():
    rankfilter = ImageFilter.RankFilter(1, 2)

    assert rankfilter.size == 1
    assert rankfilter.rank == 2


def test_builtinfilter_p():
    builtinFilter = ImageFilter.BuiltinFilter()

    with pytest.raises(ValueError):
        builtinFilter.filter(hopper("P"))


def test_kernel_not_enough_coefficients():
    with pytest.raises(ValueError):
        ImageFilter.Kernel((3, 3), (0, 0))


def test_consistency_3x3():
    with Image.open("Tests/images/hopper.bmp") as source:
        with Image.open("Tests/images/hopper_emboss.bmp") as reference:
            kernel = ImageFilter.Kernel(
                (3, 3),
                # fmt: off
                (-1, -1,  0,
                 -1,  0,  1,
                 0,   1,  1),
                # fmt: on
                0.3,
            )
            source = source.split() * 2
            reference = reference.split() * 2

            for mode in ["L", "LA", "RGB", "CMYK"]:
                assert_image_equal(
                    Image.merge(mode, source[: len(mode)]).filter(kernel),
                    Image.merge(mode, reference[: len(mode)]),
                )


def test_consistency_5x5():
    with Image.open("Tests/images/hopper.bmp") as source:
        with Image.open("Tests/images/hopper_emboss_more.bmp") as reference:
            kernel = ImageFilter.Kernel(
                (5, 5),
                # fmt: off
                (-1, -1, -1, -1,  0,
                 -1, -1, -1,  0,  1,
                 -1, -1,  0,  1,  1,
                 -1,  0,  1,  1,  1,
                 0,   1,  1,  1,  1),
                # fmt: on
                0.3,
            )
            source = source.split() * 2
            reference = reference.split() * 2

            for mode in ["L", "LA", "RGB", "CMYK"]:
                assert_image_equal(
                    Image.merge(mode, source[: len(mode)]).filter(kernel),
                    Image.merge(mode, reference[: len(mode)]),
                )
